%% Copyright (C) 2016 Wildfire Games.
%% This file is part of 0 A.D.
%%
%% 0 A.D. is free software: you can redistribute it and/or modify
%% it under the terms of the GNU General Public License as published by
%% the Free Software Foundation, either version 2 of the License, or
%% (at your option) any later version.
%%
%% 0 A.D. is distributed in the hope that it will be useful,
%% but WITHOUT ANY WARRANTY; without even the implied warranty of
%% MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
%% GNU General Public License for more details.
%%
%% You should have received a copy of the GNU General Public License
%% along with 0 A.D.  If not, see <http://www.gnu.org/licenses/>.

-module(mod_gamelist).

-behaviour(gen_mod).

% Hack to fix build. Ejabberd rewrote the module build system without leaving any
% instructions on how to upgrade. This is from Stack Overflow
% (http://stackoverflow.com/questions/32542102/ejabberd-debugging)
%TODO: Proper version checking
%-ifndef(LAGER).
%-define(LAGER, 1).
%-endif.
%-ifndef(NO_EXT_LIB).
%-define(NO_EXT_LIB, 1).
%-endif.

%% Handy shortcuts to improve readability
-record(xmlelement, {name = ""     :: string(),
                     attrs = []    :: [{string(), string()}],
                     children = [] :: [{xmlcdata, iodata()} | xmlelement()]}).
-type xmlelement() :: #xmlelement{}.
% TODO: Evaluate if we actually need dbentry
-record(dbentry, {key :: jid,
                  value :: xmlelement()}).

-include("ejabberd.hrl").
-include("jlib.hrl").

-export([start/2, stop/1, process_gamelist_iq/3]).

-define(GAMELIST_NS, "jabber:iq:gamelist").

start(Host, _Opts) ->
	?INFO_MSG("Starting mod_gamelist", []),
	% Create a table named 'games' with the default options
	mnesia:create_table(games,
		[{attributes, record_info(fields, dbentry)}, {record_name, dbentry}]),
	%% Make sure that the table is empty, as create_table won't overwrite an existing
	%% table
	mnesia:clear_table(games),
	%% Handle all stanzas addressed to the server containing the jabberd:iq:gamelist
	%% namespace with process_gamelist_iq/3 and queue pending stanzas. May be
	%% interesting to experiment with using parallel execution instead of queued.
	gen_iq_handler:add_iq_handler(ejabberd_local, Host, ?GAMELIST_NS, ?MODULE,
		process_gamelist_iq, one_queue).

stop(Host) ->
	?INFO_MSG("Stopping mod_gamelist", []),
	mnesia:delete_table(games),
	gen_iq_handler:remove_iq_handler(ejabberd_local, Host, ?GAMELIST_NS).

% Handle gamelist IQ stanzas
process_gamelist_iq(From, To, IQ) ->
	?INFO_MSG("Gamelist stanza received from ~p: ~n ~p ~n", [From, IQ]),
	% TODO: Add a checked room field to the message, keep per room databases
	case IQ#iq.type of
		set ->
			% Find and extract the command
			CommandXml = lists:keyfind("command", 2, IQ#iq.sub_el#xmlelement.children),
			[{xmlcdata, Command}] = CommandXml#xmlelement.children,
			?INFO_MSG("Command: ~p~n", [Command]),
			% TODO: Something about changestate
			case Command of
				<<"register">> ->
					% Extract game
					Game = lists:keyfind("game", 2, IQ#iq.sub_el#xmlelement.children),
					% Get host's (sender's) IP so that client know who to connect to
					Origin = ejabberd_sm:get_user_ip(From#jid.user, From#jid.server, From#jid.resource),
					HostIPString = inet_parse:ntoa(element(1, Origin)),
					% Merge in the IP
					MergedGameProperties = lists:keyreplace("ip", 1, Game#xmlelement.attrs, {"ip", HostIPString}),
					MergedGame = Game#xmlelement{attrs = MergedGameProperties},
					?INFO_MSG("Attempting to register game: ~p~nFrom: ~p", [MergedGame, From]),
					% Save the game to the database
					mnesia:dirty_write(games, {dbentry, From, MergedGame}),
					% TODO: Multicast new gamelist
					?INFO_MSG("'dev' users: ~p", [p1_fsm:sync_send_all_state_event(mnesia:dirty_read(muc_online_room, {"dev", "localhost"}), {get_disco_item, From, "en"}, 100)]),%mod_muc_admin:get_room_occupants("dev", "localhost")]),
					Room = #jid{user = "dev", server = "localhost", resource = "", luser = "dev", lserver = "localhost", lresource = ""},
					ejabberd_router:route(To, Room, jlib:iq_to_xml(get_gamelist_result_iq())),
					% Aknowledge registration
					IQ#iq{type = result, sub_el = []};
				<<"unregister">> ->
					?INFO_MSG("Attempting to unregister game...", []),
					mnesia:dirty_delete(games, From),
					IQ#iq{type = result, sub_el = []};
				_Else ->
					IQ#iq{type = error, sub_el = [IQ#iq.sub_el, ?ERR_NOT_ACCEPTABLE]}
			end;
		get ->
			% Dump database 
			?INFO_MSG("Attempting to send game listings...", []),
			get_gamelist_result_iq();
		_Else ->
			IQ#iq{type = error, sub_el = [IQ#iq.sub_el, ?ERR_NOT_ACCEPTABLE]}
	end.

get_gamelist_result_iq() ->
	%% Create a Query List Comprehension handle for the database, and
	%% evaluate it to get all the database contents
	{atomic, Games} = mnesia:transaction(fun() -> qlc:eval(mnesia:table(games)) end),
	?INFO_MSG("Recieved: ~p from DB.~n", [Games]),
	GamesXml = lists:map(fun(Game) -> Game#dbentry.value end, Games),
	SubEl = #xmlelement{name = <<"query">>, attrs = [{<<"xmlns">>, ?GAMELIST_NS}], children = GamesXml},
	#iq{type = result, xmlns = ?GAMELIST_NS, sub_el = [SubEl]}.
